import os
import yaml
import logging
import pkg_resources

log = logging.getLogger(__name__)

from .lib.signature import BadDNSSignature
from .lib.errors import BadDNSSignatureException


class BadDNS_base:
    def __init__(
        self,
        target,
        http_client_class=None,
        dns_client=None,
        signatures_dir=None,
        custom_nameservers=None,
        cli=False,
        **kwargs,
    ):
        self.target = self.set_target(target)
        self.http_client_class = http_client_class
        self.dns_client = dns_client
        self.signatures_dir = signatures_dir
        self.signatures = []
        self.load_signatures(signatures_dir)
        self.custom_nameservers = custom_nameservers
        self.parent_class = kwargs.get("parent_class", "self")
        self.cli = cli

    # hook to allow external manipulation of target assignment
    def set_target(self, target):
        return target

    def infomsg(self, msg):
        if self.cli:
            log.info(msg)
        else:
            log.debug(msg)

    def load_signatures(self, signatures_dir=None):
        if signatures_dir:
            if not os.path.exists(signatures_dir):
                raise BadDNSSignatureException(f"Signatures directory [{signatures_dir}] does not exist")
        else:
            signatures_dir = pkg_resources.resource_filename("baddns", "signatures")

        log.debug(f"attempting to load signatures from: {signatures_dir}")

        for filename in os.listdir(signatures_dir):
            if filename.endswith(".yml"):
                file_path = os.path.join(signatures_dir, filename)

                # Open each file and load the YAML contents
                try:
                    with open(file_path, "r") as file:
                        signature_data = yaml.safe_load(file)
                        signature = BadDNSSignature()
                        signature.initialize(**signature_data)
                        self.signatures.append(signature)
                except BadDNSSignatureException as e:
                    log.error(f"Error loading signature from {filename}: {e}")
        if len(self.signatures) == 0:
            raise BadDNSSignatureException(f"No signatures were successfuly loaded from [{signatures_dir}]")
        else:
            log.debug(f"Loaded [{str(len(self.signatures))}] signatures from [{signatures_dir}]")


def get_all_modules(*args, **kwargs):
    return [m for m in BadDNS_base.__subclasses__()]



from baddns.base import BadDNS_base
from .errors import BadDNSFindingException
import logging

log = logging.getLogger(__name__)


class Finding:
    def __init__(self, finding_dict):
        self.finding_dict = {}
        target = finding_dict.get("target", None)
        self.finding_dict["target"] = target
        if not target:
            raise BadDNSFindingException("Target is required in a Finding")

        description = finding_dict.get("description", "N/A")
        if not isinstance(description, str):
            raise BadDNSFindingException("description field must be a str")
        self.finding_dict["description"] = description

        confidence = finding_dict.get("confidence", None)
        if confidence == None or confidence not in ["CONFIRMED", "PROBABLE", "POSSIBLE", "UNLIKELY"]:
            raise BadDNSFindingException(
                "Confidence must be present and must be one of: CONFIRMED, PROBABLE, POSSIBLE, UNLIKELY"
            )
        self.finding_dict["confidence"] = confidence

        signature = finding_dict.get("signature", None)
        if not signature:
            raise BadDNSFindingException("signature is required in a Finding")
        self.finding_dict["signature"] = signature

        indicator = finding_dict.get("indicator", None)
        if not indicator:
            raise BadDNSFindingException("indicator is required in a Finding")
        self.finding_dict["indicator"] = indicator

        trigger = finding_dict.get("trigger", None)
        if not indicator:
            raise BadDNSFindingException("trigger is required in a Finding")
        if isinstance(trigger, list):
            trigger = ", ".join(trigger)
        elif isinstance(trigger, str):
            pass
        else:
            raise BadDNSFindingException("trigger must be either str or list")
        self.finding_dict["trigger"] = trigger

        module = finding_dict.get("module", None)
        if not module:
            raise BadDNSFindingException("Module is required in a Finding")
        if not issubclass(module, BadDNS_base):
            raise BadDNSFindingException("Module was not a valid baddns module")
        self.finding_dict["module"] = module.name

        found_domains = finding_dict.get("found_domains", None)
        if found_domains:
            self.finding_dict["found_domains"] = found_domains

    def to_dict(self):
        return self.finding_dict

    def __str__(self):
        return str(self.finding_dict)



import re
import logging
import dns.asyncresolver

log = logging.getLogger(__name__)


class DNSManager:
    dns_record_types = ["A", "AAAA", "MX", "CNAME", "NS", "SOA", "TXT", "NSEC"]

    _dns_name_regex = r"(?:\w(?:[\w-]{0,100}\w)?\.)+[^\W_]{1,63}"
    dns_name_regex = re.compile(_dns_name_regex, re.I)

    def __init__(self, target, dns_client=None, custom_nameservers=None):
        if not dns_client:
            self.dns_client = dns.asyncresolver.Resolver()
        else:
            self.dns_client = dns_client

        if custom_nameservers:
            self.dns_client.nameservers = custom_nameservers

        self.tld_nameservers = None

        self.target = target
        self.reset_answers()
        self.ips = []

    def reset_answers(self):
        self.answers = {key: None for key in self.dns_record_types}
        self.answers.update({"NoAnswer": False, "NXDOMAIN": False})

    @staticmethod
    def get_ipv4(a_records):
        ipv4 = []
        for answer in a_records:
            log.debug(f"Found IPV4 address: {answer}")
            ipv4.append(str(answer))
        return ipv4

    @staticmethod
    def get_ipv6(aaaa_records):
        ipv6 = []
        for answer in aaaa_records:
            log.debug(f"Found IPV6 address: {answer}")
            ipv6.append(str(answer))
        return ipv6

    # These were stolen from BBOT. Thanks TheTechromancer for saving me from hours of suffering.
    @staticmethod
    def _clean_dns_record(record):
        if not isinstance(record, str):
            record = str(record.to_text())
        return str(record).rstrip(".").lower()

    def process_answer(self, answer, rdatatype):
        results = set()

        if answer == None:
            return []

        for record in answer:
            """
            Extract whatever hostnames/IPs a DNS records points to
            """

            rdtype = str(record.rdtype.name).upper()
            if rdtype in ("A", "AAAA", "NS", "CNAME", "PTR"):
                results.add(self._clean_dns_record(record))
            elif rdtype == "SOA":
                results.add(self._clean_dns_record(record.mname))
            elif rdtype == "MX":
                results.add(self._clean_dns_record(record.exchange))
            elif rdtype == "SRV":
                results.add(self._clean_dns_record(record.target))
            elif rdtype == "TXT":
                for s in record.strings:
                    s = s.decode()
                    results.add(s)
            elif rdtype == "NSEC":
                results.add(self._clean_dns_record(record.next))
            else:
                log.warning(f'Unknown DNS record type "{rdtype}"')
        return list(results)

    async def do_resolve(self, target, rdatatype):
        try:
            r = self.process_answer(await self.dns_client.resolve(target, rdatatype), rdatatype)
        except (dns.resolver.NoAnswer, dns.resolver.NoNameservers) as e:
            log.debug(f"encountered error with dns_client.resolve(): {e}")
            self.answers["NoAnswer"] = True
            return
        except dns.resolver.NXDOMAIN:
            self.answers["NXDOMAIN"] = True
            return
        except dns.resolver.LifetimeTimeout as e:
            log.debug(f"Dns Timeout: {e}")
            return
        if r and len(r) > 0:
            if rdatatype == "A":
                self.ips.extend(self.get_ipv4(r))
            elif rdatatype == "AAAA":
                self.ips.extend(self.get_ipv6(r))

            elif rdatatype == "CNAME":
                cname_chain = []

                while 1:
                    result_cname = r[0]
                    cname_chain.append(result_cname)
                    target = result_cname
                    try:
                        r = self.process_answer(await self.dns_client.resolve(target, "CNAME"), "CNAME")
                        if len(r) == 0:
                            break
                    except (dns.resolver.NoAnswer, dns.resolver.NXDOMAIN) as e:
                        log.debug(f"Error resolving cname chain: {e}")
                        break
                return cname_chain
            return r

    async def dispatchDNS(self, omit_types=[]):
        log.debug(f"attempting to resolve {self.target}")
        log.debug(f"dispatching DNS with the following nameservers: {' '.join(self.dns_client.nameservers)}")
        for rdatatype in self.dns_record_types:
            if rdatatype in omit_types:
                continue
            try:
                self.answers[rdatatype] = await self.do_resolve(self.target, rdatatype)
            except dns.resolver.LifetimeTimeout:
                log.debug(f"Got LifetimeTimeout for rdatatype [{rdatatype}] for target [{self.target}]")
                self.answers[rdatatype] = None



import logging

from .errors import BadDNSSignatureException

log = logging.getLogger(__name__)


class BadDNSSignature:
    validModes = ["http", "dns_nxdomain", "dns_nosoa"]
    validSources = ["dnsreaper", "nucleitemplates", "self"]

    def __init__(self):
        self.signature = {
            "service_name": None,
            "source": None,
            "identifiers": {"cnames": [], "not_cnames": [], "ips": [], "nameservers": []},
            "mode": None,
            "matcher_rule": {},
        }

    def initialize(self, **kwargs):
        self.signature["mode"] = kwargs.get("mode", None)
        self.signature["source"] = kwargs.get("source", None)
        self.signature["service_name"] = kwargs.get("service_name", None)
        identifiers = kwargs.get("identifiers", {})
        self.signature["identifiers"] = {}
        self.signature["identifiers"]["cnames"] = identifiers.get("cnames", [])
        self.signature["identifiers"]["not_cnames"] = identifiers.get("not_cnames", [])
        self.signature["identifiers"]["ips"] = identifiers.get("ips", [])
        self.signature["identifiers"]["nameservers"] = identifiers.get("nameservers", [])
        self.signature["matcher_rule"] = kwargs.get("matcher_rule", None)

        if not self.signature["mode"]:
            raise BadDNSSignatureException(f"mode is a required attribute")

        if self.signature["mode"] not in self.validModes:
            raise BadDNSSignatureException(f"Supplied mode [{self.signature.mode}] in not a valid mode")

        if not self.signature["source"]:
            raise BadDNSSignatureException(f"source is a required attribute")

        if self.signature["source"] not in self.validSources:
            raise BadDNSSignatureException(f"Supplied mode [{self.signature['source']}] is not a valid mode")

        if not self.signature["service_name"]:
            raise BadDNSSignatureException(f"service_name is a required attribute")

        if self.signature["mode"] == "http":
            if not self.signature["matcher_rule"]:
                raise BadDNSSignatureException(f"http mode requires a matcher_rule entry")

        if self.signature["mode"].startswith("dns"):
            if self.signature["matcher_rule"]:
                raise BadDNSSignatureException(f"In dns modes, matcher_rule should not be set")

        if self.signature["mode"] == "dns_nosoa":
            if len(self.signature["identifiers"]["nameservers"]) == 0:
                raise BadDNSSignatureException(f"In dns_nosoa mode, nameservers are required")

    def output(self):
        return self.signature

    def summarize_matcher_rule(self):
        summary = []

        if "matchers" in self.signature["matcher_rule"].keys():
            for matcher in self.signature["matcher_rule"]["matchers"]:
                if matcher["type"] == "word":
                    words = ", ".join(matcher["words"])
                    condition = matcher.get("condition", "")
                    part = matcher.get("part", "")
                    summary.append(f"[Words: {words} | Condition: {condition} | Part: {part}]")
            return ", ".join(summary) + f" Matchers-Condition: {self.signature['matcher_rule']['matchers-condition']}"
        else:
            return "No matchers in signature"



import os
import sys
import yaml
import json
import string
import random
import httpx
import dns.resolver

rand_pool = string.ascii_lowercase

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(SCRIPT_DIR))

from lib.signature import BadDNSSignature
from lib.matcher import Matcher
from lib.errors import BadDNSSignatureException


def dns_request(domain):
    try:
        answers = dns.resolver.resolve(domain, "A")
        return [answer.to_text() for answer in answers]

    except dns.resolver.NXDOMAIN:
        return "NXDOMAIN"

    except dns.resolver.Timeout:
        return "Timeout"

    except dns.resolver.NoAnswer:
        return "NoAnswer"

    except dns.resolver.NoNameservers:
        return "NoNameservers"

    except Exception as e:
        return str(e)


testsig = """
identifiers:
  cnames:
  - helpscoutdocs.com
  ips: []
  nameservers: []
matcher_rule:
  matchers:
  - condition: or
    part: body
    type: word
    words:
    - Not Found
  matchers-condition: and
mode: http
service_name: helpscoutdocs.com
source: dnsreaper
"""


def rand_string(length=12):
    return "".join([random.choice(rand_pool) for _ in range(int(length))])


yaml_data = yaml.safe_load(testsig)
match_table = {}


headers = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36",
    "Accept-Language": "en-US,en;q=0.9",
    "Cache-Control": "no-cache",
    "Connection": "keep-alive",
    "Pragma": "no-cache",
    "Upgrade-Insecure-Requests": "1",
}


def process_file(file_path):
    error = None
    signature_pass = False

    with open(file_path, "r") as file:
        sig_to_test = file.read()

    sig = BadDNSSignature()
    sig_yaml = yaml.safe_load(sig_to_test)
    try:
        sig.initialize(**sig_yaml)
    except BadDNSSignatureException as e:
        return False, {}, f"Failed Signature Validation: [{e}]"

    if sig.signature["mode"] == "http":
        matcher_rule = sig.signature["matcher_rule"]
        matcher = Matcher(matcher_rule)
        if len(sig.signature["identifiers"]["cnames"]) > 0:
            for cname_dict in sig.signature["identifiers"]["cnames"]:
                if cname_dict["type"] == "word":
                    cname = cname_dict["value"]
                    match_found = False
                    for scheme in ("http", "https"):
                        for follow_redirects in [True, False]:
                            try:
                                url = f"{scheme}://{rand_string()}.{cname}"
                                r = httpx.get(url, headers=headers, follow_redirects=follow_redirects, timeout=5)
                                if matcher.is_match(r):
                                    match_found = True
                            except httpx.ConnectError:
                                pass
                    if match_found:
                        signature_pass = True
                        match_table[cname] = True
                    else:
                        match_table[cname] = False
                else:
                    pass
                    # TODO: Support other types
        else:
            signature_pass = True
        if signature_pass == False:
            error = "No CNAMES passed random subdomain matcher validation"

    elif sig.signature["mode"] == "dns_nxdomain":
        for cname_dict in sig.signature["identifiers"]["cnames"]:
            if cname_dict["type"] == "word":
                cname = cname_dict["value"]
                test_domain = f"{rand_string()}.{cname}"
                r = dns_request(test_domain)
                if r == "NXDOMAIN":
                    signature_pass = True
                    match_table[cname] = True
                else:
                    match_table[cname] = False
        if signature_pass == False:
            error = "No CNAMES gave expected NXDOMAIN response"

    elif sig.signature["mode"] == "dns_nosoa":
        signature_pass = True

    return signature_pass, match_table, error


def main():
    if len(sys.argv) != 2:
        print("Usage: python script.py <input_file>")
        sys.exit(1)

    file_path = sys.argv[1]
    signature_pass, match_table, error = process_file(file_path)

    # Convert results to JSON and print
    result = {
        "signature_pass": signature_pass,
        "match_table": match_table,
        "error": error,
    }
    print(json.dumps(result))


if __name__ == "__main__":
    main()



from baddns.base import BadDNS_base
from baddns.lib.findings import Finding
from baddns.lib.dnsmanager import DNSManager

import logging
import dns.zone
import dns.query
import asyncio

log = logging.getLogger(__name__)


class BadDNS_zonetransfer(BadDNS_base):
    name = "zonetransfer"
    description = "Attempt a DNS zone transfer"
    zone_records = []
    zone_nameservers = []

    def __init__(self, target, **kwargs):
        super().__init__(target, **kwargs)
        self.target_dnsmanager = DNSManager(
            target, dns_client=self.dns_client, custom_nameservers=self.custom_nameservers
        )

    def parse_zone(self, zone):
        for name, node in zone.nodes.items():
            for rdataset in node.rdatasets:
                record_type = dns.rdatatype.to_text(rdataset.rdtype)
                for rdata in rdataset:
                    raw_name = name.to_text()
                    if str(raw_name) == "@":
                        processed_name = self.target_dnsmanager.target
                    else:
                        processed_name = f"{raw_name}.{self.target_dnsmanager.target}"
                    record = (processed_name, record_type, rdata.to_text())
                    if record[0] not in self.zone_records:
                        self.zone_records.append(record[0])

    async def zone_transfer(self, nameserver, domain):
        ns_ips = await self.target_dnsmanager.do_resolve(nameserver, "A")
        if ns_ips:
            ns_ip = ns_ips[0]
            log.debug(f"resolved nameserver [{nameserver}] ip: [{ns_ip}]")
            try:
                zone = await asyncio.to_thread(dns.zone.from_xfr, dns.query.xfr(ns_ip, domain, lifetime=6, timeout=6))

            except (TimeoutError, dns.exception.Timeout):
                log.debug("TimeoutError attempting zone transfer")
                return False
            except ConnectionResetError:
                log.debug("ConnectionResetError attempting zone transfer")
                return False
            except dns.xfr.TransferError as e:
                log.debug(f"{nameserver} ({ns_ip}): {e}")
                return False
            except EOFError:
                log.debug("EOFError attempting zone transfer")
                return False
            except dns.exception.FormError:
                log.debug("dns.exception.FormError attempting zone transfer")
            self.zone_nameservers.append(nameserver)
            self.parse_zone(zone)
            return True

    async def dispatch(self):
        zone_transfer_detected = False
        await self.target_dnsmanager.dispatchDNS(omit_types=["A", "MX", "AAAA", "CNAME", "SOA", "TXT", "NSEC"])
        if self.target_dnsmanager.answers["NS"]:
            for ns in self.target_dnsmanager.answers["NS"]:
                log.debug(f"Attempting Zone Transfer against NS [{ns}] for target [{self.target_dnsmanager.target}]")
                r = await self.zone_transfer(ns, self.target_dnsmanager.target)
                if r:
                    zone_transfer_detected = True
                    log.info(
                        f"Successful Zone Transfer against NS [{ns}] for target [{self.target_dnsmanager.target}]"
                    )
        return zone_transfer_detected

    def analyze(self):
        findings = []
        if self.zone_records:
            findings.append(
                Finding(
                    {
                        "target": self.target_dnsmanager.target,
                        "description": "Successful Zone Transfer",
                        "confidence": "CONFIRMED",
                        "signature": "N/A",
                        "indicator": "Successful XFR Request",
                        "trigger": self.zone_nameservers,
                        "module": type(self),
                        "found_domains": self.zone_records,
                    }
                )
            )
        return findings



import tldextract

from baddns.base import BadDNS_base

from baddns.lib.dnsmanager import DNSManager
from baddns.lib.httpmanager import HttpManager
from baddns.lib.whoismanager import WhoisManager
from baddns.lib.matcher import Matcher
from baddns.lib.findings import Finding

import logging

log = logging.getLogger(__name__)


class BadDNS_cname(BadDNS_base):
    name = "CNAME"
    description = "Check for dangling CNAME records and interrogate them for subdomain takeover opportunities"

    def __init__(self, target, **kwargs):
        super().__init__(target, **kwargs)

        self.direct_mode = kwargs.get("direct_mode", False)
        self.target_dnsmanager = DNSManager(
            target, dns_client=self.dns_client, custom_nameservers=self.custom_nameservers
        )
        self.target_httpmanager = None
        self.cname_dnsmanager = None
        self.cname_whoismanager = None

    async def dispatch(self):
        await self.target_dnsmanager.dispatchDNS()
        if self.direct_mode == False:
            if self.target_dnsmanager.answers["CNAME"] != None:
                self.infomsg(
                    f"Found CNAME(S) [{' -> '.join([self.target_dnsmanager.target] + self.target_dnsmanager.answers['CNAME'])}]"
                )
                self.subject = self.target_dnsmanager.answers["CNAME"][-1]
            else:
                if self.parent_class == "self":
                    self.infomsg("No CNAME Found :/")
                return False
        else:
            log.debug("Direct mode enabled. Target will be checked for takeover instead of target's CNAME")
            self.subject = self.target
        self.cname_dnsmanager = DNSManager(self.subject, dns_client=self.dns_client)
        await self.cname_dnsmanager.dispatchDNS(omit_types=["CNAME", "NSEC"])

        # if the domain resolves, we can try for HTTP connections
        if not self.cname_dnsmanager.answers["NXDOMAIN"]:
            log.debug("CNAME resolved correctly, proceeding with HTTP dispatch")
            self.target_httpmanager = HttpManager(self.target, http_client_class=self.http_client_class)
            await self.target_httpmanager.dispatchHttp()
            log.debug("HTTP dispatch complete")
        # if the cname doesn't resolve, we still need to see if the base domain is unregistered
        # even if it is registered, we still use whois to check for expired domains
        log.debug("performing WHOIS lookup")

        self.cname_whoismanager = WhoisManager(self.subject)
        await self.cname_whoismanager.dispatchWHOIS()
        log.debug("WHOIS dispatch complete")
        return True

    def analyze(self):
        findings = []
        if self.direct_mode == True:
            trigger = ["self"]
        else:
            trigger = self.target_dnsmanager.answers["CNAME"]
        if self.cname_dnsmanager.answers["NXDOMAIN"]:
            signature_match = False
            indicator = None

            self.infomsg(f"Got NXDOMAIN for CNAME {self.cname_dnsmanager.target}. Checking against signatures...")
            for sig in self.signatures:
                if sig.signature["mode"] == "dns_nxdomain":
                    log.debug(f"Trying signature {sig.signature['service_name']}")
                    sig_cnames = [c["value"] for c in sig.signature["identifiers"]["cnames"]]
                    for sig_cname in sig_cnames:
                        log.debug(f"Checking CNAME {self.cname_dnsmanager.target} against {sig_cname}")
                        if self.cname_dnsmanager.target.endswith(sig_cname):
                            signature_match = True
                            log.debug(f"CNAME {self.cname_dnsmanager.target} vulnerable ({sig_cname})")
                            indicator = sig_cname
                            findings.append(
                                Finding(
                                    {
                                        "target": self.target_dnsmanager.target,
                                        "description": f"Dangling CNAME, probable subdomain takeover (NXDOMAIN technique)",
                                        "confidence": "PROBABLE",
                                        "signature": sig.signature["service_name"],
                                        "indicator": indicator,
                                        "trigger": trigger,
                                        "module": type(self),
                                    }
                                )
                            )
                            break
            if (
                signature_match == False
                and trigger[-1] != "self"
                and tldextract.extract(trigger[-1]).registered_domain
                != tldextract.extract(self.target_dnsmanager.target).registered_domain
            ):
                findings.append(
                    Finding(
                        {
                            "target": self.target_dnsmanager.target,
                            "description": f"Dangling CNAME, possible subdomain takeover (NXDOMAIN technique)",
                            "confidence": "POSSIBLE",
                            "signature": "GENERIC",
                            "indicator": "Generic Dangling CNAME",
                            "trigger": trigger,
                            "module": type(self),
                        }
                    )
                )
            else:
                log.debug(
                    f"Not reporting generic cname for trigger [{trigger}] from domain [{self.target_dnsmanager.target}]"
                )

        else:
            log.debug("Starting HTTP analysis")

            http_results = [
                self.target_httpmanager.http_allowredirects_results,
                self.target_httpmanager.http_denyredirects_results,
                self.target_httpmanager.https_allowredirects_results,
                self.target_httpmanager.https_denyredirects_results,
            ]

            for sig in self.signatures:
                if sig.signature["mode"] == "http":
                    log.debug(f"Trying signature {sig.signature['service_name']}")
                    if len(sig.signature["identifiers"]["cnames"]) > 0:
                        log.debug(
                            f"Signature contains cnames [{sig.signature['identifiers']['cnames']}], checking them"
                        )
                        if not any(
                            cname_dict["value"] in self.subject
                            for cname_dict in sig.signature["identifiers"]["cnames"]
                        ):
                            log.debug(f"no match for {sig.signature['identifiers']['cnames']} for in {self.subject}")
                            continue
                        log.debug("passed CNAME check")

                    if len(sig.signature["identifiers"]["ips"]) > 0:
                        log.debug(f"Signature contains ips [{sig.signature['identifiers']['ips']}], checking them")
                        if not any(
                            ip_signature in self.cname_dnsmanager.ips
                            for ip_signature in sig.signature["identifiers"]["ips"]
                        ):
                            log.debug(
                                f"no match for {sig.signature['identifiers']['ips']} for in {self.cname_dnsmanager.ips}"
                            )
                            continue
                        log.debug("passed IPS")

                    m = Matcher(sig.signature)
                    log.debug("Checking for HTTP matches")
                    if any(m.is_match(hr) for hr in http_results if hr is not None):
                        log.debug(f"CNAME {self.cname_dnsmanager.target} Vulnerable")
                        log.debug(f"With matcher_rule {sig.signature['matcher_rule']}")
                        findings.append(
                            Finding(
                                {
                                    "target": self.target_dnsmanager.target,
                                    "description": f"Dangling CNAME, probable subdomain takeover (HTTP String Match)",
                                    "confidence": "PROBABLE",
                                    "signature": sig.signature["service_name"],
                                    "indicator": sig.summarize_matcher_rule(),
                                    "trigger": trigger,
                                    "module": type(self),
                                }
                            )
                        )

        # check whois data for unregistered and expiring domains
        if self.cname_whoismanager.whois_result:
            for whois_finding in self.cname_whoismanager.analyzeWHOIS():
                findings.append(
                    Finding(
                        {
                            "target": self.target_dnsmanager.target,
                            "description": f"CNAME {whois_finding}",
                            "confidence": "CONFIRMED",
                            "signature": "N/A",
                            "indicator": "Whois Data",
                            "trigger": self.subject,
                            "module": type(self),
                        }
                    )
                )

        return findings



#!/usr/bin/env python3
import re
import os
import ast
import sys
import ipaddress
import yaml
import logging
from abc import ABC, abstractmethod

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(SCRIPT_DIR))

from lib.signature import BadDNSSignature
from lib.errors import BadDNSSignatureException

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
handler = logging.FileHandler("readsources.log")
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)


class Transformer(ABC):
    @abstractmethod
    def transform(self, input_data):
        pass

    def writeSignature(self, shortname, signatureName):
        output_directory = "signatures_to_test"
        os.makedirs(output_directory, exist_ok=True)  # Creates the directory if it does not exist
        output_file_path = os.path.join(output_directory, f"{shortname}_{signatureName}.yml")
        with open(output_file_path, "w") as f:
            signature_candidate = BadDNSSignature()
            signature_data = self.map_values()
            logger.info(f"Got signature data: [{signature_data}]")
            try:
                signature_candidate.initialize(**signature_data)
            except BadDNSSignatureException as e:
                logger.info(f"Error encountered validating signature data: [{e}]")

            logger.info(
                f"Signature data validated. Final signature: \n****************************{yaml.dump(signature_candidate.output())}\n****************************"
            )
            logger.info(f"writing to: [{output_file_path}]")
            yaml.dump(signature_candidate.output(), f)

    def _is_ip_address(self, value):
        try:
            ipaddress.IPv4Address(value)
            return True
        except ipaddress.AddressValueError:
            pass

        try:
            ipaddress.IPv6Address(value)
            return True
        except ipaddress.AddressValueError:
            pass

        return False


class NucleiTemplatesTransformer(Transformer):
    def transform(self, signature_data):
        self.shortname = "nucleitemplates"
        self.data = {"source": self.shortname}
        self.nucleiconvert()

    def parse_dsl(self, dsl_rule):
        # contains host regex
        host_regex = re.compile(r'\!contains\(host,"(?P<domain>.+?)\"\)')
        host_regex_match = host_regex.match(dsl_rule)
        if host_regex_match:
            return host_regex_match.group("domain"), "to_cname"
        return None, None

    def map_values(self):
        values = {
            "service_name": data["info"]["name"],
            "source": "nucleitemplates",
            "identifiers": {"cnames": [], "not_cnames": [], "ips": [], "nameservers": []},
            "mode": None,
            "matcher_rule": None,
        }

        if "http" in data.keys():
            values["mode"] = "http"

            http_data = data["http"][0]
            if "matchers" in http_data.keys():
                matcher_rule = {
                    "matchers-condition": http_data.get("matchers-condition", "and"),
                    "matchers": [],
                }
                for matcher in http_data["matchers"]:
                    # Ignore unknown types
                    if matcher["type"] not in ["status", "word", "regex", "dsl"]:
                        continue

                    if matcher["type"] == "word":
                        matcher["part"] = matcher.get("part", "body")

                        if matcher["part"] == "host":
                            negative = matcher.get("negative", False)

                            for w in matcher["words"]:
                                if negative:
                                    values["identifiers"]["not_cnames"].append({"type": "word", "value": w})
                                else:
                                    values["identifiers"]["cnames"].append({"type": "word", "value": w})
                            continue

                        matcher["condition"] = matcher.get("condition", "and")
                    if matcher["type"] == "dsl":
                        for dsl_rule in matcher["dsl"]:
                            parsed_dsl, dsl_type = self.parse_dsl(dsl_rule)

                            if dsl_type == "to_cname":
                                values["identifiers"]["cnames"].append({"type": "word", "value": parsed_dsl})

                    if matcher["type"] == "status":
                        matcher["status"] = matcher["status"][0]

                    matcher_rule["matchers"].append(matcher)
                values["matcher_rule"] = matcher_rule
        elif "dns" in data.keys():
            values["mode"] = "dns_nxdomain"
            dns_data = data["dns"][0]
            if "matchers" in dns_data.keys():
                for matcher in dns_data["matchers"]:
                    if matcher["type"] == "word":
                        for word in matcher.get("words", []):
                            if word != "NXDOMAIN":
                                if self._is_ip_address(word):
                                    values["identifiers"]["ips"].append({"type": "word", "value": word})
                                else:
                                    values["identifiers"]["cnames"].append({"type": "word", "value": word})
                    elif matcher["type"] == "regex":
                        for regex_str in matcher.get("regex", []):
                            values["identifiers"]["cnames"].append(
                                {"type": "regex", "value": regex_str.replace("CNAME\t", "")}
                            )
        return values


class DnsReaperSignatureTransformer(Transformer):
    use_case_to_mode_mapping = {
        "cname_found_but_string_in_body": "http",
        "cname_found_but_status_code": "http",
        "cname_or_ip_found_but_string_in_body": "http",
        "ip_found_but_string_in_body": "http",
        "cname_found_but_NX_DOMAIN": "dns_nxdomain",
        "ns_found_but_no_SOA": "dns_nosoa",
    }

    def transform(self, signature_data):
        self.shortname = "dnsreaper"
        self.data = {"source": self.shortname}
        self.variables = {}
        self._visit(ast.parse(signature_data))

    def map_values(self):
        values = {}
        identifiers = {"cnames": [], "not_cnames": [], "ips": [], "nameservers": []}

        for key, value in self.data.items():
            if key == "http_strings" and value:
                if isinstance(value, list):
                    for http_string in value:
                        values.setdefault("matcher_rule", {"matchers-condition": "and", "matchers": []})[
                            "matchers"
                        ].append({"type": "word", "words": [http_string], "condition": "or", "part": "body"})
                else:
                    values.setdefault("matcher_rule", {"matchers-condition": "and", "matchers": []})[
                        "matchers"
                    ].append({"type": "word", "words": [value], "condition": "or", "part": "body"})
            elif key == "status_code" and value is not None:
                values.setdefault("matcher_rule", {"matchers-condition": "and", "matchers": []})["matchers"].append(
                    {"type": "status", "status": int(value)}
                )

            elif key in ["ips", "nameservers", "cnames", "not_cnames"]:
                identifiers[key] = value

            elif key == "use_case":
                values["mode"] = self.use_case_to_mode_mapping.get(value, value)
            else:
                values[key] = value

        values["identifiers"] = identifiers
        return values

    def _visit(self, node):
        method = "_visit_" + node.__class__.__name__
        visitor = getattr(self, method, self._visit_generic)
        visitor(node)

        for child in ast.iter_child_nodes(node):
            self._visit(child)

    def _visit_generic(self, node):
        pass

    def _visit_Assign(self, node):
        if len(node.targets) > 0 and isinstance(node.targets[0], ast.Name):
            target_var_name = node.targets[0].id
            if isinstance(node.value, ast.List) and all(isinstance(elt, ast.Constant) for elt in node.value.elts):
                self.variables[target_var_name] = [elt.value for elt in node.value.elts]
            elif isinstance(node.value, ast.Constant):
                self.variables[target_var_name] = [node.value.value]

    def _visit_List(self, node):
        ips = [elt.s for elt in node.elts if isinstance(elt, ast.Str) and self._is_ip_address(elt.s)]
        if ips:
            if "ips" in self.data.keys():
                self.data["ips"] += ips
            else:
                self.data["ips"] = ips

    def _visit_Call(self, node):
        call = node
        self.data["use_case"] = (
            call.func.id
            if isinstance(call.func, ast.Name)
            else (call.func.attr if isinstance(call.func, ast.Attribute) else None)
        )
        self.data["service_name"] = next(
            (kw.value.s for kw in call.keywords if kw.arg == "service" and isinstance(kw.value, ast.Str)), None
        )
        self.data["http_strings"] = [
            kw.value.s
            for kw in call.keywords
            if kw.arg == "domain_not_configured_message" and isinstance(kw.value, ast.Str)
        ]
        self.data["https"] = next(
            (kw.value.id for kw in call.keywords if kw.arg == "https" and isinstance(kw.value, ast.Name)), None
        )

        for kw in call.keywords:
            if kw.arg == "cname":
                if isinstance(kw.value, ast.Str):
                    self.data["cnames"] = [{"type": "word", "value": kw.value.s.replace("cname.", "").lstrip(".")}]

                elif isinstance(kw.value, ast.List):
                    self.data["cnames"] = [
                        {"type": "word", "value": elt.s.replace("cname.", "").lstrip(".")} for elt in kw.value.elts
                    ]

                elif isinstance(kw.value, ast.Name):
                    self.data["cnames"] = [
                        {"type": "word", "value": val.replace("cname.", "").lstrip(".")}
                        for val in self.variables.get(kw.value.id, [])
                    ]

        if call.args:  # Check for positional arguments
            for arg in call.args:
                if isinstance(arg, ast.List):
                    self.data["nameservers"] = [elt.s for elt in arg.elts]

        ns_val = next((kw.value for kw in call.keywords if kw.arg == "ns"), None)
        if ns_val:
            if isinstance(ns_val, ast.Str):
                self.data["nameservers"] = [ns_val.s]
            elif isinstance(ns_val, ast.List):
                self.data["nameservers"] = [elt.s for elt in ns_val.elts]
        self.data["sample_nameserver"] = next(
            (kw.value.s for kw in call.keywords if kw.arg == "sample_ns" and isinstance(kw.value, ast.Str)), None
        )
        self.data["status_code"] = next(
            (kw.value.n for kw in call.keywords if kw.arg == "code" and isinstance(kw.value, ast.Num)), None
        )

        for arg in call.args:
            if isinstance(arg, ast.List):
                self._visit_List(arg)


directory = "./dnsReaper/signatures"
logger.info("readsources init")
logger.info(f"Starting dnsReaper ingest, reading from: {os.path.join(os.getcwd(), directory)}")

files = os.listdir(directory)

for filename in files:
    logger.info(f"loading dnsReaper signature [{filename}]")
    if not filename.startswith("_") and filename.endswith(".py"):
        dnsReaper_transformer = DnsReaperSignatureTransformer()
        with open(f"{directory}/{filename}") as f:
            dnsReaper_transformer.transform(f.read())
            dnsReaper_transformer.writeSignature("dnsreaper", filename.split(".")[0])


directory_http = "./nuclei-templates/http/takeovers"
directory_dns = "./nuclei-templates/dns"

logger.info(
    f"Starting nuclei-templates ingest, reading from: [{os.path.join(os.getcwd(), directory_http)}] and [{os.path.join(os.getcwd(), directory_dns)}]"
)

files_http = os.listdir(os.path.expanduser(directory_http))
files_dns = os.listdir(os.path.expanduser(directory_dns))

files = []
for filename in files_http:
    files.append(os.path.join(directory_http, filename))
for filename in files_dns:
    files.append(os.path.join(directory_dns, filename))

for filepath in files:
    if "-takeover" in filepath and filepath.endswith(".yaml"):
        logger.info(f"loading nuclei-template [{filename}]")
        with open(filepath, "r") as file:
            data = yaml.safe_load(file)
        nucleitemplates_transformer = NucleiTemplatesTransformer()
        nucleitemplates_transformer.writeSignature("nucleitemplates", filepath.split("/")[-1].split(".")[0])

logger.info("readsources complete")
